import base64
import inspect
from pathlib import Path
from collections import OrderedDict
from typing import Optional, List, Dict, Any

import numpy as np
import pandas as pd
import plotly.express as px
from dash import html, dcc, Input, Output, Dash
import webviz_core_components as wcc

from .. import WebvizPluginABC, WebvizSettings, EncodedFile
from ..webviz_store import webvizstore
from ..common_cache import CACHE


# pylint: disable=too-many-arguments
class TablePlotter(WebvizPluginABC):
    """Adds a plotter to the webviz instance, using tabular data from a provided csv file.
If feature is requested, the data could also come from a database.

---

* **`csv_file`:** Path to the csv file containing the tabular data. \
                  Either absolute path or relative to the configuration file.
* **`plot_options`:** A dictionary of plot options to initialize the plot with.
* **`filter_cols`:** Dataframe columns that can be used to filter data.
* **`filter_defaults`:** A dictionary with column names as keys, \
                         and a list of column values that should be preselected in the filter. \
                         If a columm is not defined, all values are preselected for the column.
* **`column_color_discrete_maps`:** A dictionary with column names as keys, \
                                    each key containing a new dictionary with the columns \
                                    unique values as keys, and the color they should be \
                                    plotted with as value. Hex values needs quotes '' \
                                    to not be read as a comment.
* **`lock`:** If `True`, only the plot is shown, \
              all dropdowns for changing plot options are hidden.
"""

    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        csv_file: Path,
        plot_options: dict = None,
        filter_cols: list = None,
        filter_defaults: dict = None,
        column_color_discrete_maps: dict = None,
        lock: bool = False,
    ) -> None:

        super().__init__()

        self.plot_options = plot_options if plot_options else {}
        self.lock = lock
        self.csv_file = csv_file
        self.data = get_data(self.csv_file)
        self.set_filters(filter_cols)
        self.columns = list(self.data.columns)
        self.numeric_columns = list(
            self.data.select_dtypes(include=[np.number]).columns
        )
        self.filter_defaults = filter_defaults
        self.column_color_discrete_maps = column_color_discrete_maps
        self.plotly_theme = webviz_settings.theme.plotly_theme
        self.set_callbacks(app)

    def set_filters(self, filter_cols: Optional[list]) -> None:
        self.filter_cols = []
        self.use_filter = False
        if filter_cols:
            for col in filter_cols:
                if col in self.data.columns:
                    if self.data[col].nunique() != 1:
                        self.filter_cols.append(col)
            if self.filter_cols:
                self.use_filter = True

    def add_webvizstore(self) -> List[tuple]:
        return [(get_data, [{"csv_file": self.csv_file}])]

    @property
    def plots(self) -> dict:
        """A dict of available plots and their options"""
        return {
            "scatter": ["x", "y", "size", "color", "facet_col"],
            "histogram": [
                "x",
                "color",
                "facet_col",
                "barmode",
                "barnorm",
                "histnorm",
            ],
            "bar": ["x", "y", "color", "facet_col", "barmode"],
            "pie": ["values", "names"],
            "scatter_3d": ["x", "y", "z", "size", "color"],
            "line": ["x", "y", "color", "line_group", "facet_col"],
            "line_3d": ["x", "y", "z", "color"],
            "box": ["x", "y", "color", "facet_col"],
            "violin": ["x", "y", "color", "facet_col"],
            "scatter_matrix": ["dimensions", "size", "color"],
            "parallel_coordinates": ["dimensions"],
            "parallel_categories": ["dimensions"],
            "density_contour": ["x", "y", "color", "facet_col"],
        }

    @property
    def plot_args(self) -> dict:
        """A dict of possible plot options and their default values"""
        return OrderedDict(
            {
                "x": {
                    "options": self.columns,
                    "value": self.plot_options.get("x", self.columns[0]),
                    "multi": False,
                    "clearable": False,
                },
                "y": {
                    "options": self.columns,
                    "value": self.plot_options.get("y", self.columns[0]),
                    "multi": False,
                    "clearable": False,
                },
                "z": {
                    "options": self.columns,
                    "value": self.plot_options.get("z", self.columns[0]),
                    "multi": False,
                    "clearable": False,
                },
                "values": {
                    "options": self.numeric_columns,
                    "value": self.plot_options.get("values", self.numeric_columns[0]),
                    "multi": False,
                    "clearable": True,
                },
                "names": {
                    "options": self.columns,
                    "value": self.plot_options.get("names", None),
                    "multi": False,
                    "clearable": True,
                },
                "size": {
                    "options": self.numeric_columns,
                    "value": self.plot_options.get("size", None),
                    "multi": False,
                    "clearable": True,
                },
                "color": {
                    "options": self.columns,
                    "value": self.plot_options.get("color", None),
                    "multi": False,
                    "clearable": True,
                },
                "facet_col": {
                    "options": self.columns,
                    "value": self.plot_options.get("facet_col", None),
                    "multi": False,
                    "clearable": True,
                },
                "line_group": {
                    "options": self.columns,
                    "value": self.plot_options.get("line_group", None),
                    "multi": False,
                    "clearable": True,
                },
                "barmode": {
                    "options": ["stack", "group", "overlay", "relative"],
                    "value": self.plot_options.get("barmode", "stack"),
                    "multi": False,
                    "clearable": True,
                },
                "barnorm": {
                    "options": ["fraction", "percent"],
                    "value": self.plot_options.get("barnorm", None),
                    "multi": False,
                    "clearable": True,
                },
                "histnorm": {
                    "options": [
                        "percent",
                        "propability",
                        "density",
                        "propability density",
                    ],
                    "value": self.plot_options.get("histnorm", None),
                    "multi": False,
                    "clearable": True,
                },
                "trendline": {
                    "options": self.numeric_columns,
                    "value": None,
                    "multi": False,
                    "clearable": True,
                },
                "dimensions": {
                    "options": self.columns,
                    "value": self.plot_options.get("dimensions", self.columns),
                    "multi": True,
                    "clearable": True,
                },
            }
        )

    def filter_layout(self) -> Optional[list]:
        """Makes dropdowns for each dataframe column used for filtering."""
        if not self.use_filter:
            return None
        df = self.data
        dropdowns = [html.H4("Set filters")]
        for col in self.filter_cols:
            if df[col].dtype in [np.float64, np.int64]:
                min_val = df[col].min()
                max_val = df[col].max()
                mean_val = df[col].mean()
                dropdowns.append(
                    html.Div(
                        children=[
                            html.Details(
                                open=True,
                                children=[
                                    html.Summary(col.lower().capitalize()),
                                    dcc.RangeSlider(
                                        id=self.uuid(f"filter-{col}"),
                                        min=min_val,
                                        max=max_val,
                                        step=(max_val - min_val) / 10,
                                        marks={
                                            min_val: f"{min_val:.2f}",
                                            mean_val: f"{mean_val:.2f}",
                                            max_val: f"{max_val:.2f}",
                                        },
                                        value=[min_val, max_val],
                                    ),
                                ],
                            )
                        ]
                    )
                )
            else:
                elements = list(self.data[col].unique())
                dropdowns.append(
                    html.Div(
                        children=[
                            html.Details(
                                open=True,
                                children=[
                                    html.Summary(col.lower().capitalize()),
                                    wcc.Select(
                                        id=self.uuid(f"filter-{col}"),
                                        options=[
                                            {"label": i, "value": i} for i in elements
                                        ],
                                        value=elements
                                        if self.filter_defaults is None
                                        else [
                                            element
                                            for element in self.filter_defaults.get(
                                                col, elements
                                            )
                                            if element in elements
                                        ],
                                        size=min(15, len(elements)),
                                    ),
                                ],
                            )
                        ]
                    )
                )
        return dropdowns

    def plot_option_layout(self) -> List[html.Div]:
        """Renders a dropdown widget for each plot option"""
        divs = []
        # The plot type dropdown is handled separate
        divs.append(
            html.Div(
                style=self.style_options_div,
                children=[
                    html.H4("Set plot options"),
                    html.P("Plot type"),
                    dcc.Dropdown(
                        id=self.uuid("plottype"),
                        clearable=False,
                        options=[{"label": i, "value": i} for i in self.plots],
                        value=self.plot_options.get("type", "scatter"),
                    ),
                ],
            )
        )
        # Looping through all available plot options
        # and renders a dropdown widget
        for key, arg in self.plot_args.items():
            divs.append(
                html.Div(
                    style=self.style_options_div_hidden,
                    id=self.uuid(f"div-{key}"),
                    children=[
                        html.P(key),
                        dcc.Dropdown(
                            id=self.uuid(f"dropdown-{key}"),
                            clearable=arg["clearable"],
                            options=[{"label": i, "value": i} for i in arg["options"]],
                            value=arg["value"],
                            multi=arg["multi"],
                        ),
                    ],
                )
            )
        return divs

    @property
    def style_options_div(self) -> Dict[str, str]:
        """Style for active plot options"""
        return {"display": "grid"}

    @property
    def style_options_div_hidden(self) -> Dict[str, str]:
        """Style for hidden plot options"""
        return {"display": "none"}

    @property
    def layout(self) -> html.Div:
        return html.Div(
            children=[
                wcc.FlexBox(
                    children=[
                        html.Div(
                            id=self.uuid("selector-row"),
                            style={"display": "none"}
                            if self.lock
                            else {"width": "15%"},
                            children=self.plot_option_layout(),
                        ),
                        wcc.Graph(
                            id=self.uuid("graph-id"),
                            style={"height": "80vh", "width": "60%"},
                        ),
                        html.Div(style={"width": "15%"}, children=self.filter_layout()),
                    ],
                )
            ]
        )

    @property
    def plot_output_callbacks(self) -> List[Output]:
        """Creates list of output dependencies for callback
        The outputs are the graph, and the style of the plot options"""
        outputs = []
        outputs.append(Output(self.uuid("graph-id"), "figure"))
        for plot_arg in self.plot_args.keys():
            outputs.append(Output(self.uuid(f"div-{plot_arg}"), "style"))
        return outputs

    @property
    def plot_input_callbacks(self) -> List[Input]:
        """Creates list of input dependencies for callback
        The inputs are the plot type and the current value
        for each plot option
        """
        inputs = []
        inputs.append(Input(self.uuid("plottype"), "value"))
        for plot_arg in self.plot_args.keys():
            inputs.append(Input(self.uuid(f"dropdown-{plot_arg}"), "value"))
        for filtcol in self.filter_cols:
            inputs.append(Input(self.uuid(f"filter-{filtcol}"), "value"))
        return inputs

    def set_callbacks(self, app: Dash) -> None:
        @app.callback(self.plugin_data_output, self.plugin_data_requested)
        def _user_download_data(data_requested: Optional[int]) -> Optional[EncodedFile]:
            return (
                {
                    "filename": "table-plotter.csv",
                    "content": base64.b64encode(
                        get_data(self.csv_file).to_csv().encode()
                    ).decode("ascii"),
                    "mime_type": "text/csv",
                }
                if data_requested
                else None
            )

        @app.callback(self.plot_output_callbacks, self.plot_input_callbacks)
        def _update_output(*args: Any) -> tuple:
            """Updates the graph and shows/hides plot options"""
            plot_type = args[0]
            # pylint: disable=protected-access
            plotfunc = getattr(px._chart_types, plot_type)
            plotargs = {}
            div_style = []
            data = self.data
            # Filter dataframe if filter columns are available
            if self.use_filter:
                plot_inputs = args[1 : -len(self.filter_cols)]
                filter_inputs = args[-len(self.filter_cols) :]
                data = filter_dataframe(data, self.filter_cols, filter_inputs)
            else:
                plot_inputs = args[1:]
            for name, plot_arg in zip(self.plot_args.keys(), plot_inputs):
                if plot_type in ["parallel_coordinates"] and name == "dimensions":
                    # This plot type only accepts numerical data
                    plot_arg = [val for val in plot_arg if val in self.numeric_columns]
                if name in self.plots[plot_type]:
                    plotargs[name] = plot_arg
                    div_style.append(self.style_options_div)

                    if (
                        name == "color"
                        and self.column_color_discrete_maps is not None
                        and plot_arg in self.column_color_discrete_maps
                        and "color_discrete_map"
                        in inspect.signature(plotfunc).parameters
                    ):
                        plotargs[
                            "color_discrete_map"
                        ] = self.column_color_discrete_maps.get(plot_arg)
                else:
                    div_style.append(self.style_options_div_hidden)
            return (plotfunc(data, template=self.plotly_theme, **plotargs), *div_style)


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def get_data(csv_file: Path) -> pd.DataFrame:
    return pd.read_csv(csv_file, index_col=None)


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def filter_dataframe(
    dframe: pd.DataFrame, columns: list, column_values: List[list]
) -> pd.DataFrame:
    df = dframe.copy()
    if not isinstance(columns, list):
        columns = [columns]
    for filt, col in zip(column_values, columns):
        if isinstance(filt, list):
            if df[col].dtype in [np.float64, np.int64]:
                df = df.loc[df[col].between(filt[0], filt[1])]
            else:
                df = df.loc[df[col].isin(filt)]
        else:
            df = df.loc[df[col] == filt]
    return df



from pathlib import Path
from typing import List

import pandas as pd
from dash import dash_table

from .. import WebvizPluginABC
from ..webviz_store import webvizstore
from ..common_cache import CACHE


class DataTable(WebvizPluginABC):
    """Adds a table to the webviz instance, using tabular data from a provided csv file.
If feature is requested, the data could also come from a database.

---

* **`csv_file`:** Path to the csv file containing the tabular data. Either absolute \
              path or relative to the configuration file.
* **`sorting`:** If `True`, the table can be sorted interactively based \
             on data in the individual columns.
* **`filtering`:** If `True`, the table can be filtered based on values in the \
               individual columns.
* **`pagination`:** If `True`, only a subset of the table is displayed at once. \
                Different subsets can be viewed from 'previous/next' buttons
"""

    def __init__(
        self,
        csv_file: Path,
        sorting: bool = True,
        filtering: bool = True,
        pagination: bool = True,
    ):

        super().__init__()

        self.csv_file = csv_file
        self.df = get_data(self.csv_file)
        self.sorting = sorting
        self.filtering = filtering
        self.pagination = pagination

    def add_webvizstore(self) -> List[tuple]:
        return [(get_data, [{"csv_file": self.csv_file}])]

    @property
    def layout(self) -> dash_table.DataTable:
        return dash_table.DataTable(
            columns=[{"name": i, "id": i} for i in self.df.columns],
            data=self.df.to_dict("records"),
            sort_action="native" if self.sorting else "none",
            filter_action="native" if self.filtering else "none",
            page_action="native" if self.pagination else "none",
        )


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def get_data(csv_file: Path) -> pd.DataFrame:
    return pd.read_csv(csv_file)



import re
import warnings
from typing import Any, Dict, Iterable, Optional, Tuple

try:
    # Python 3.8+
    from typing import TypedDict
    from importlib.metadata import requires, version, PackageNotFoundError
except ImportError:
    # Python < 3.8
    from typing_extensions import TypedDict
    from importlib_metadata import requires, version, PackageNotFoundError  # type: ignore


class PluginProjectMetaData(TypedDict):
    dist_version: str
    dependencies: Dict[str, str]
    documentation_url: Optional[str]
    download_url: Optional[str]
    source_url: Optional[str]
    tracker_url: Optional[str]


def _plugin_dist_dependencies(plugin_dist_name: str) -> Dict[str, str]:
    """Returns overview of all dependencies (indirect + direct) of a given
    plugin project installed in the current environment.

    Key is package name of dependency, value is (installed) version string.
    """

    untraversed_dependencies = set([plugin_dist_name])
    requirements = {}

    while untraversed_dependencies:
        sub_dependencies = requires(untraversed_dependencies.pop())

        if sub_dependencies is None:
            continue

        for sub_dependency in sub_dependencies:
            split = re.split(r"[;<>~=()]", sub_dependency, 1)
            package_name = split[0].strip().replace("_", "-").lower()

            if package_name not in requirements:
                # Only include package in dependency list
                # if it is not an "extra" dependency...
                if len(split) == 1 or "extra" not in split[1]:
                    try:
                        # ...and if it is actually installed (there are dependencies
                        # in setup.py that e.g. are not installed on certain Python
                        # versions and operating system combinations).
                        requirements[package_name] = version(package_name)
                        untraversed_dependencies.add(package_name)
                    except PackageNotFoundError:
                        pass

    return {k: requirements[k] for k in sorted(requirements)}


def load_webviz_plugins_with_metadata(
    distributions: Iterable, loaded_plugins: Dict[str, Any]
) -> Tuple[Dict[str, dict], Dict[str, PluginProjectMetaData]]:
    """Loads the given distributions, finds entry points corresponding to webviz-config
    plugins, and put them into the mutable input dictionary loaded_plugins
    (key is plugin name string, value is reference to plugin class).
    Also returns a dictionary of plugin metadata.
    """

    plugin_project_metadata: Dict[str, PluginProjectMetaData] = {}
    plugin_metadata: Dict[str, dict] = {}

    for dist in distributions:
        for entry_point in dist.entry_points:
            if entry_point.group == "webviz_config_plugins":

                dist_name = dist.metadata["name"]

                project_urls = {
                    value.split(",")[0]: value.split(",")[1].strip()
                    for (key, value) in dist.metadata.items()
                    if key == "Project-URL"
                }

                if (
                    entry_point.name in plugin_metadata
                    and dist_name not in plugin_project_metadata
                ):
                    warnings.warn(
                        f"Multiple versions of plugin with name {entry_point.name}. Already "
                        f"loaded from project {plugin_metadata[entry_point.name]['dist_name']}. "
                        f"Overwriting using plugin with from project {dist_name}",
                        RuntimeWarning,
                    )

                if dist_name not in plugin_project_metadata:
                    plugin_project_metadata[dist_name] = PluginProjectMetaData(
                        {
                            "dist_version": dist.version,
                            "dependencies": _plugin_dist_dependencies(dist_name),
                            "documentation_url": project_urls.get("Documentation"),
                            "download_url": project_urls.get("Download"),
                            "source_url": project_urls.get("Source"),
                            "tracker_url": project_urls.get("Tracker"),
                        }
                    )

                plugin_metadata[entry_point.name] = {
                    "dist_name": dist.metadata["name"],
                }

                loaded_plugins[entry_point.name] = entry_point.load()

    return (plugin_metadata, plugin_project_metadata)



# Sets the window size of the browser (crucial in --headless mode).
from selenium.webdriver.chrome.options import Options


def pytest_setup_options():
    options = Options()
    options.add_argument("--window-size=1920,1080")
    return options



import re
import shutil
import pathlib
from typing import Optional

from tqdm import tqdm
from dash import Dash
import flask


class WebvizAssets:
    """Dash applications by default host static resources from a folder called
    ./assets, relative to the root application folder.

    In order to facilitate hot reloading, and fast building of Webviz
    applications from the configuration file, this class facilitates handling
    of static assets.

    Individual plugins can add assets to a common instance of WebvizAssets
    by calling the .add(filename) function. This adds the resource, and
    at the same time returns the resource URI which the plugin can use.

    If the webviz instance is in non-portable mode, the Flask/Dash application
    is routed to the actual location of the files, making hot reload and
    testing fast.

    When creating a portable webviz instance however, the files are copied
    over the ./resources/assets folder, and normal Dash usage applies.

    In both portable and non-portable mode, WebvizAssets makes sure there are
    no name conflicts (i.e. it supports multiple assets on different paths,
    but with same filename) and also assignes URI friendly resource IDs.
    """

    def __init__(self) -> None:
        self._assets: dict = {}
        self._portable = False

    @property
    def portable(self) -> bool:
        return self._portable

    @portable.setter
    def portable(self, portable: bool) -> None:
        self._portable = portable

    def _base_folder(self) -> str:
        return "assets" if self.portable else "temp"

    def add(self, filename: pathlib.Path) -> str:
        """Calling this function makes the filename given as input
        available as a hosted asset when the application is running.
        The returned string is a URI which the plugin optionally
        can use internally (e.g. as "src" in image elements).

        Calling this function with the same input path
        multiple times will return the same URI.

        Filenames added to WebvizAssets that ends with .css or .js
        are loaded automatically in the browser by Dash,
        both in non-portable and portable mode.
        """

        path = pathlib.Path(filename)

        if filename not in self._assets.values():
            assigned_id = self._generate_id(path.name)
            self._assets[assigned_id] = filename
        else:
            assigned_id = {v: k for k, v in self._assets.items()}[filename]

        return str(pathlib.Path(self._base_folder()) / assigned_id)

    def directly_host_assets(self, app: Dash) -> None:
        """In non-portable mode, this function can be called by the
        application. It routes the Dash application to the added assets on
        disk, making hot reloading and more interactive development of the
        application possible.
        """

        if self._portable:
            raise RuntimeError(
                "The function WebvizAssets.directly_host_assets() "
                "method is only meaningful in a non-portable settings."
            )

        @app.server.route(f"/{self._base_folder()}/<path:asset_id>")
        def _send_file(asset_id: str) -> Optional[flask.wrappers.Response]:
            if asset_id in self._assets:  # Only serve white listed resources
                path = pathlib.Path(self._assets[asset_id])
                return flask.send_from_directory(str(path.parent), path.name)
            return None

        # Add .css and .js files to auto-loaded Dash assets
        for asset_id, asset_path in self._assets.items():
            if asset_path.suffix == ".css":
                app.config.external_stylesheets.append(
                    f"./{self._base_folder()}/{asset_id}"
                )
            elif asset_path.suffix == ".js":
                app.config.external_scripts.append(
                    f"./{self._base_folder()}/{asset_id}"
                )

    def make_portable(self, asset_folder: pathlib.Path) -> None:
        """Copy over all added assets to the given folder (asset_folder)."""

        for assigned_id, filename in tqdm(
            self._assets.items(),
            bar_format="{l_bar} {bar} | Copied {n_fmt}/{total_fmt}",
        ):
            tqdm.write(f"Copying over {filename}")
            shutil.copyfile(filename, asset_folder / assigned_id)

    def _generate_id(self, filename: str) -> str:
        """From the filename, create a safe resource id not already present"""
        asset_id = base_id = re.sub(
            "[^-a-z0-9._]+", "", filename.lower().replace(" ", "_")
        )

        count = 1
        while asset_id in self._assets:
            count += 1
            asset_id = f"{base_id}{count}"

        return asset_id


WEBVIZ_ASSETS = WebvizAssets()



import dash

from webviz_config.common_cache import CACHE
from webviz_config.generic_plugins import _example_plugin


def test_example_plugin(dash_duo):

    app = dash.Dash(__name__)
    app.config.suppress_callback_exceptions = True
    CACHE.init_app(app.server)
    title = "Example"
    page = _example_plugin.ExamplePlugin(app, title)
    app.layout = page.layout
    dash_duo.start_server(app)
    btn = dash_duo.find_element("#" + page.uuid("submit-button"))
    assert btn.text == "Submit"
    text = dash_duo.find_element("#" + page.uuid("output-state"))
    assert text.text == "Button has been pressed 0 times."
    btn.click()
    dash_duo.wait_for_contains_text(
        "#" + page.uuid("output-state"), "Button has been pressed 1 times", timeout=2
    )
    assert text.text == "Button has been pressed 1 times."



import copy
import inspect
import pathlib
from typing import Callable, List, Dict


class SharedSettingsSubscriptions:
    """The user can configure common settings, shared between different cointainers,
    under a key called `shared_settings` in the configuration file. Since it originates
    from a native yaml file, the content is dictionaries/strings/ints/floats/dates.

    Third-party plugin packages might want to check early if the `shared_settings`
    they use are reasonable, and/or do some transformations on them.
    """

    def __init__(self) -> None:
        self._subscriptions: List[Dict] = []

    def subscribe(self, key: str) -> Callable:
        """This is the decorator, which third-party plugin packages will use."""

        def register(function: Callable) -> Callable:
            self._subscriptions.append({"key": key, "function": function})
            return function

        return register

    def transformed_settings(
        self, shared_settings: Dict, config_folder: str, portable: bool
    ) -> Dict:
        """Called from the app template, which returns the `shared_settings`
        after all third-party package subscriptions have done their
        (optional) transfomrations.
        """
        shared_settings = copy.deepcopy(shared_settings)

        for subscription in self._subscriptions:
            key, function = subscription["key"], subscription["function"]
            shared_settings[key] = function(
                *[
                    pathlib.Path(config_folder)
                    if arg == "config_folder"
                    else portable
                    if arg == "portable"
                    else shared_settings.get(arg)
                    for arg in inspect.getfullargspec(function).args
                ]
            )

        return shared_settings


SHARED_SETTINGS_SUBSCRIPTIONS = SharedSettingsSubscriptions()



from pathlib import Path
from typing import List

from dash import html

from .. import WebvizPluginABC
from ..webviz_assets import WEBVIZ_ASSETS


class BannerImage(WebvizPluginABC):
    """Adds a full width banner image, with an optional overlayed title.
Useful on e.g. the front page for introducing a field or project.

---

* **`image`:** Path to the picture you want to add. \
               Either absolute path or relative to the configuration file.
* **`title`:** Title which will be overlayed over the banner image.
* **`color`:** Color to be used for the font.
* **`shadow`:** Set to `False` if you do not want text shadow for the title.
* **`height`:** Height of the banner image (in pixels).
"""

    TOOLBAR_BUTTONS: List[str] = []

    def __init__(
        self,
        image: Path,
        title: str = "",
        color: str = "white",
        shadow: bool = True,
        height: int = 300,
    ):

        super().__init__()

        self.image = image
        self.title = title
        self.color = color
        self.shadow = shadow
        self.height = height

        self.image_url = WEBVIZ_ASSETS.add(image)

    @property
    def layout(self) -> html.Div:

        style = {
            "color": self.color,
            "backgroundImage": f"url({self.image_url})",
            "height": f"{self.height}px",
        }

        if self.shadow:
            style["textShadow"] = "0.05em 0.05em 0"

            if self.color == "white":
                style["textShadow"] += " rgba(0, 0, 0, 0.7)"
            else:
                style["textShadow"] += " rgba(255, 255, 255, 0.7)"

        return html.Div(self.title, className="_banner_image", style=style)



import os
import secrets

import flask

from ._is_reload_process import is_reload_process
from ._oauth2 import Oauth2


class LocalhostToken:
    """Uses a method similar to jupyter notebook. This method is only used during
    interactive usage on localhost, and the workflow is as follows:

    - During the flask app building, a one-time-token (ott) and a cookie_token
      is generated.
    - When the app is ready, the user needs to "login" using this
      one-time-token in the url (http://localhost:{port}?ott={token})
    - If ott is valid - a cookie with a separate token is set, and the
      one-time-token is discarded. The cookie is then used for subsequent
      requests.

    If the user fails both providing a valid one-time-token and a valid cookie
    token, all localhost requests gets a 401.

    If the app is in non-portable mode, the one-time-token and
    cookie token  are reused on app reload (in order to ensure live reload
    works without requiring new login).

    The port is used as a postfix on the cookie name in order to make sure that
    two different localhost applications running simultaneously do not interfere.
    """

    def __init__(self, app: flask.app.Flask, port: int, oauth2: Oauth2 = None):
        self._app = app
        self._port = port
        self._oauth2 = oauth2

        if not is_reload_process():
            # One time token (per run) user has to provide
            # when visiting the localhost app the first time.
            self._ott = os.environ["WEBVIZ_OTT"] = LocalhostToken.generate_token()

            # This is the cookie token set in the users browser after
            # successfully providing the one time token
            self._cookie_token = os.environ[
                "WEBVIZ_COOKIE_TOKEN"
            ] = LocalhostToken.generate_token()

        else:
            self._ott = os.environ["WEBVIZ_OTT"]
            self._cookie_token = os.environ["WEBVIZ_COOKIE_TOKEN"]

        self._ott_validated = False
        self.set_request_decorators()

    @staticmethod
    def generate_token() -> str:
        return secrets.token_urlsafe(nbytes=64)

    @property
    def one_time_token(self) -> str:
        return self._ott

    def set_request_decorators(self) -> None:
        # pylint: disable=inconsistent-return-statements
        @self._app.before_request
        def _check_for_ott_or_cookie():  # type: ignore[no-untyped-def]

            if not self._ott_validated and self._ott == flask.request.args.get("ott"):
                self._ott_validated = True
                flask.g.set_cookie_token = True  # pylint: disable=assigning-non-slot
                return flask.redirect(flask.request.base_url)

            if self._cookie_token == flask.request.cookies.get(
                f"cookie_token_{self._port}"
            ):
                self._ott_validated = True

                if self._oauth2:
                    return self._oauth2.check_access_token()

            else:
                flask.abort(401)

        @self._app.after_request
        def _set_cookie_token_in_response(
            response: flask.wrappers.Response,
        ) -> flask.wrappers.Response:
            if flask.g.get("set_cookie_token", False):
                response.set_cookie(
                    key=f"cookie_token_{self._port}", value=self._cookie_token
                )
            return response



import copy
from typing import Dict, Any, Mapping

from ._theme_class import WebvizConfigTheme


class WebvizSettings:
    """This class contains global Webviz settings that will be made available
    to all plugins through the special argument named webviz_settings.
    """

    def __init__(self, shared_settings: Dict[str, Any], theme: WebvizConfigTheme):
        if not isinstance(shared_settings, dict):
            raise TypeError("shared_settings must be of type dict")

        if not isinstance(theme, WebvizConfigTheme):
            raise TypeError("theme must be of type WebvizConfigTheme")

        self._shared_settings = shared_settings
        self._theme = theme

    @property
    def shared_settings(self) -> Mapping[str, Any]:
        return copy.deepcopy(self._shared_settings)

    @property
    def theme(self) -> WebvizConfigTheme:
        return copy.deepcopy(self._theme)



try:
    # Python 3.8+
    from importlib.metadata import version, PackageNotFoundError
except ModuleNotFoundError:
    # Python < 3.8
    from importlib_metadata import version, PackageNotFoundError  # type: ignore

from ._theme_class import WebvizConfigTheme
from ._webviz_settings_class import WebvizSettings
from ._localhost_token import LocalhostToken
from ._is_reload_process import is_reload_process
from ._plugin_abc import WebvizPluginABC, EncodedFile, ZipFileMember
from ._shared_settings_subscriptions import SHARED_SETTINGS_SUBSCRIPTIONS
from .webviz_instance_info import WEBVIZ_INSTANCE_INFO
from ._oauth2 import Oauth2

try:
    __version__ = version("webviz-config")
except PackageNotFoundError:
    # package is not installed
    pass



import json
import time
import shutil
import pathlib
import subprocess


def binary_available() -> bool:
    """Returns True if the Radix CLI binary (rx) is
    availabile in $PATH, otherwise returns False.
    """
    return shutil.which("rx") is not None


def _trigger_token_aquisition(update_only: bool = False) -> None:
    # There exist no pure login command in rx as of April 2021
    # See https://github.com/equinor/radix-cli/issues/20.
    # pylint: disable=subprocess-run-check
    subprocess.run(
        [
            "rx",
            "get",
            "application",
            "--application",
            "dummy_name",
            "--context",
            "dummy_context",
        ],
        timeout=15 if update_only else None,
        stdout=subprocess.PIPE if update_only else None,
        stderr=subprocess.PIPE if update_only else None,
        # capture_output=update_only <-- Added in Python 3.7
    )


def logged_in() -> bool:

    _trigger_token_aquisition(update_only=True)

    config = pathlib.Path.home() / ".radix" / "config"

    if config.is_file():
        token_expiry_time = json.loads(config.read_text())["sessionConfig"]["expiresOn"]
        if token_expiry_time > time.time():
            return True

    return False


def log_in() -> None:
    _trigger_token_aquisition()


def application_exists(application_name: str, context: str) -> bool:
    result = subprocess.run(
        [
            "rx",
            "get",
            "application",
            "--application",
            application_name,
            "--context",
            context,
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        # capture_output=True,  <-- Added in Python 3.7
        check=True,
    )
    return not result.stderr


def create_application(
    application_name: str,
    owner_email: str,
    repository_url: str,
    shared_secret: str,
    wbs: str,
    context: str,
) -> str:
    result = subprocess.run(
        [
            "rx",
            "create",
            "application",
            "--application",
            application_name,
            "--owner",
            owner_email,
            "--config-branch",
            "main",
            "--repository",
            repository_url,
            "--shared-secret",
            shared_secret,
            "--wbs",
            wbs,
            "--context",
            context,
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        # capture_output=True,
        check=True,
    )

    stderr = result.stderr.decode()
    if not stderr.startswith("ssh-rsa"):
        error_message = stderr
        if "registerApplicationBadRequest" in stderr:
            error_message += (
                f"Is {repository_url} being used by another Radix application?"
            )
        raise RuntimeError(error_message)

    return stderr


def build_and_deploy_application(application_name: str, context: str) -> None:
    subprocess.run(
        [
            "rx",
            "create",
            "job",
            "build-deploy",
            "--application",
            application_name,
            "--branch",
            "main",
            "--context",
            context,
        ],
        check=True,
    )


def set_radix_secret(
    application_name: str,
    environment: str,
    component: str,
    key: str,
    value: str,
    context: str,
) -> None:

    max_radix_build_time = 60 * 60
    sleep_per_attempt = 10

    for _ in range(max_radix_build_time // sleep_per_attempt):
        try:
            subprocess.run(
                [
                    "rx",
                    "set",
                    "environment-secret",
                    "--await-reconcile",
                    "--application",
                    application_name,
                    "--environment",
                    environment,
                    "--component",
                    component,
                    "--secret",
                    key,
                    "--value",
                    value,
                    "--context",
                    context,
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                # capture_output=True, <-- First vailable in Python 3.7
                check=True,
            )
            return
        except subprocess.CalledProcessError:
            time.sleep(sleep_per_attempt)

    raise RuntimeError("Failed setting Radix secret. Is Radix still building?")



from ._localhost_open_browser import LocalhostOpenBrowser
from ._available_port import get_available_port
from ._silence_flask_startup import silence_flask_startup
from ._dash_component_utils import calculate_slider_step
from ._deprecate_webviz_settings_attribute_in_dash_app import (
    deprecate_webviz_settings_attribute_in_dash_app,
)



import os
import time
import urllib
import warnings
import threading
import webbrowser

from .._is_reload_process import is_reload_process
from .._user_preferences import get_user_preference
from . import terminal_colors


class LocalhostOpenBrowser:
    # pylint: disable=too-few-public-methods

    def __init__(self, port: int, token: str):
        self._port = port
        self._token = token

        if not is_reload_process():
            # Only open new browser tab if not a reload process
            threading.Thread(target=self._timer).start()

    def _timer(self) -> None:
        """Waits until the app is ready, and then opens the page
        in the default browser.
        """

        timeout = 120  # maximum number of seconds to wait before timeout

        for _ in range(timeout):
            if self._app_ready():
                self._open_new_tab()
                return

            time.sleep(1)

        print(
            f"WARNING: Webviz application still not ready after {timeout}s.\n"
            "Will not open browser automatically. Your private one-time login link:\n"
            f"{self._url(with_token=True)}"
        )

    def _url(self, with_token: bool = False) -> str:
        return (
            f"http://localhost:{self._port}"
            + f"{'?ott=' + self._token if with_token else ''}"
        )

    @staticmethod
    def _get_browser_controller() -> webbrowser.BaseBrowser:

        if get_user_preference("browser") is not None:
            try:
                return webbrowser.get(using=get_user_preference("browser"))
            except webbrowser.Error:
                warnings.warn("Could not find the user preferred browser.")

        for browser in ["firefox", "chrome", "chromium-browser"]:
            try:
                return webbrowser.get(using=browser)
            except webbrowser.Error:
                pass

        # Return default browser if none of the
        # preferred browsers are installed:
        return webbrowser.get()

    def _app_ready(self) -> bool:
        """Check if the flask instance is ready."""

        no_proxy_env = os.environ.get("NO_PROXY")
        os.environ["NO_PROXY"] = "localhost"

        try:
            with urllib.request.urlopen(self._url()):  # nosec
                app_ready = True
        except urllib.error.HTTPError:
            # The flask instance responded with some HTTP error (likely 401),
            # but is otherwise ready to accept connections
            app_ready = True
        except urllib.error.URLError:
            # The flask instance has not started
            app_ready = False
        except ConnectionResetError:
            # The flask instance has started but (correctly) abort
            # request due to "401 Unauthorized"
            app_ready = True
        finally:
            os.environ["NO_PROXY"] = no_proxy_env if no_proxy_env else ""

        return app_ready

    def _open_new_tab(self) -> None:
        """Open the url (with token) in the default browser."""

        print(
            f"{terminal_colors.GREEN}{terminal_colors.BOLD}"
            f" Opening the application ({self._url()}) in your browser.\n"
            f"{terminal_colors.BLUE}"
            " Note that your browser might display security issues.\n"
            " See: https://equinor.github.io/webviz-config/#/?id=localhost-hsts\n"
            f"{terminal_colors.GREEN}"
            " Press CTRL + C in this terminal window to stop the application."
            f"{terminal_colors.END}"
        )

        LocalhostOpenBrowser._get_browser_controller().open_new_tab(
            self._url(with_token=True)
        )



try:
    # Python 3.8+
    from importlib.metadata import entry_points
except ModuleNotFoundError:
    # Python < 3.8
    from importlib_metadata import entry_points  # type: ignore

from .. import WebvizConfigTheme
from ._default_theme import default_theme

installed_themes = {default_theme.theme_name: default_theme}

__all__ = ["installed_themes"]

for entry_point in entry_points().get("webviz_config_themes", []):
    theme = entry_point.load()

    globals()[entry_point.name] = theme
    __all__.append(entry_point.name)

    if isinstance(theme, WebvizConfigTheme):
        installed_themes[theme.theme_name] = theme



from typing import Dict, Union, Callable, Tuple, cast, Optional, List, Type
import inspect

from ._plugin_abc import WebvizPluginABC
from . import _deprecation_store as _ds


def deprecated_plugin(
    deprecation_info: str = "",
) -> Callable[[Type[WebvizPluginABC]], Type[WebvizPluginABC]]:
    def wrapper(
        original_plugin: Type[WebvizPluginABC],
    ) -> Type[WebvizPluginABC]:

        _ds.DEPRECATION_STORE.register_deprecated_plugin(
            _ds.DeprecatedPlugin(
                original_plugin,
                f"Plugin '{original_plugin.__name__}' has been deprecated.",
                deprecation_info,
            )
        )

        return original_plugin

    return wrapper


def deprecated_plugin_arguments(
    check: Union[Dict[str, str], Callable[..., Optional[Tuple[str, str]]]]
) -> Callable:
    def decorator(original_init_method: Callable) -> Callable:
        original_method_args = inspect.getfullargspec(original_init_method).args

        if callable(check):
            check_args = inspect.getfullargspec(check).args
            verified_args: List[str] = []
            for check_arg in check_args:
                for original_arg in original_method_args:
                    if check_arg == original_arg:
                        verified_args.append(check_arg)

            _ds.DEPRECATION_STORE.register_deprecated_plugin_argument(
                _ds.DeprecatedArgumentCheck(
                    original_init_method,
                    original_init_method.__name__,
                    verified_args,
                    check,
                    inspect.getsource(check),
                )
            )

        elif isinstance(check, dict):
            for original_arg in original_method_args:
                if original_arg in check.keys():
                    short_message = cast(Tuple[str, str], check[original_arg])[0]
                    long_message = cast(Tuple[str, str], check[original_arg])[1]
                    _ds.DEPRECATION_STORE.register_deprecated_plugin_argument(
                        _ds.DeprecatedArgument(
                            original_init_method,
                            original_init_method.__name__,
                            original_arg,
                            "",
                            short_message,
                            long_message,
                        )
                    )
        return original_init_method

    return decorator



import re
from pathlib import Path
from typing import List
from xml.etree import ElementTree  # nosec

import bleach
import markdown
from markdown.extensions import Extension
from markdown.inlinepatterns import ImageInlineProcessor, IMAGE_LINK_RE
from dash import dcc

from .. import WebvizPluginABC
from ..webviz_assets import WEBVIZ_ASSETS
from ..webviz_store import webvizstore


class _WebvizMarkdownExtension(Extension):
    def __init__(self, base_path: Path):
        self.base_path = base_path

        super().__init__()

    def extendMarkdown(self, md: markdown.Markdown) -> None:
        md.inlinePatterns.register(
            item=_MarkdownImageProcessor(IMAGE_LINK_RE, md, self.base_path),
            name="image_link",
            priority=50,
        )


class _MarkdownImageProcessor(ImageInlineProcessor):
    def __init__(self, image_link_re: str, md: markdown.core.Markdown, base_path: Path):
        self.base_path = base_path

        super().__init__(image_link_re, md)

    def handleMatch(self, m, data: str) -> tuple:  # type: ignore[no-untyped-def, override]
        image, start, index = super().handleMatch(m, data)

        if image is None or not image.get("title"):
            return image, start, index

        src = image.get("src")
        caption = image.get("title")

        if src is None or src.startswith("http"):
            raise ValueError(
                f"Image path {src} has been given. Only images "
                "available on the file system can be added."
            )

        image_path = Path(src)

        if not image_path.is_absolute():
            image_path = (self.base_path / image_path).resolve()

        url = WEBVIZ_ASSETS.add(image_path)

        image_style = ""
        for style_prop in image.get("alt", default="").split(","):
            prop, value = style_prop.split("=")
            if prop == "width":
                image_style += f"width: {value};"
            elif prop == "height":
                image_style += f"height: {value};"

        if image_style:
            image.set("style", image_style)

        image.set("src", url)
        image.set("class", "_markdown_image")

        container = ElementTree.Element("span")
        container.append(image)

        ElementTree.SubElement(
            container, "span", attrib={"class": "_markdown_image_caption"}
        ).text = caption

        return container, start, index


class Markdown(WebvizPluginABC):
    """Renders and includes the content from a Markdown file.

    ---

    * **`markdown_file`:** Path to the markdown file to render and include. \
                        Either absolute path or relative to the configuration file.

    ---

    Images are supported, and should in the markdown file be given as either
    relative paths to the markdown file itself, or as absolute paths.

    > The markdown syntax for images has been extended to support \
    providing width and/or height for individual images (optional). \
    To specify the dimensions write e.g.
    > ```markdown
    > ![width=40%,height=300px](./example_banner.png "Some caption")
    > ```

    """

    ALLOWED_TAGS = [
        "a",
        "b",
        "blockquote",
        "br",
        "cite",
        "code",
        "dd",
        "del",
        "details",
        "div",
        "dl",
        "dt",
        "em",
        "h1",
        "h2",
        "h3",
        "h4",
        "h5",
        "h6",
        "hr",
        "i",
        "img",
        "ins",
        "li",
        "mark",
        "ol",
        "p",
        "pre",
        "q",
        "s",
        "span",
        "strong",
        "sub",
        "summary",
        "sup",
        "table",
        "tbody",
        "td",
        "th",
        "thead",
        "tr",
        "ul",
    ]

    ALLOWED_ATTRIBUTES = {
        "*": ["id", "class", "style"],
        "a": ["href", "alt", "title"],
        "details": ["open"],
        "img": ["src", "alt", "title", "style"],
        "li": ["value"],
        "ol": ["reversed", "start", "type"],
        "td": ["colspan", "rowspan"],
        "th": ["colspan", "rowspan", "scope"],
    }

    ALLOWED_STYLES = ["width", "height"]

    def __init__(self, markdown_file: Path):

        super().__init__()

        self.markdown_file = markdown_file

        self.html = bleach.clean(
            markdown.markdown(
                get_path(self.markdown_file).read_text(),
                extensions=[
                    "fenced_code",
                    "tables",
                    "sane_lists",
                    _WebvizMarkdownExtension(base_path=markdown_file.parent),
                ],
            ),
            tags=Markdown.ALLOWED_TAGS,
            attributes=Markdown.ALLOWED_ATTRIBUTES,
            styles=Markdown.ALLOWED_STYLES,
        )

        # Workaround for upstream issue https://github.com/plotly/dash-core-components/issues/746,
        # where we convert void html tags from <tag> to <tag/>.
        self.html = re.sub("<img (.*?[^/])>", r"<img \1/>", self.html)
        self.html = self.html.replace("<br>", "<br/>").replace("<hr>", "<hr/>")

    def add_webvizstore(self) -> List[tuple]:
        return [(get_path, [{"path": self.markdown_file}])]

    @property
    def layout(self) -> dcc.Markdown:
        return dcc.Markdown(
            self.html, dangerously_allow_html=True, className="webviz-config-markdown"
        )


@webvizstore
def get_path(path: Path) -> Path:
    return path



"""These are the basic Webviz configuration plugins, distributed through
the utility itself.
"""

try:
    # Python 3.8+
    from importlib.metadata import distributions
except ModuleNotFoundError:
    # Python < 3.8
    from importlib_metadata import distributions  # type: ignore

from ._utils import load_webviz_plugins_with_metadata, PluginProjectMetaData


PLUGIN_METADATA, PLUGIN_PROJECT_METADATA = load_webviz_plugins_with_metadata(
    distributions(), globals()
)



from typing import List

from dash import html

from .. import WebvizPluginABC


class ExampleTour(WebvizPluginABC):
    @property
    def tour_steps(self) -> List[dict]:
        return [
            {"id": self.uuid("blue_text"), "content": "This is the first step"},
            {"id": self.uuid("red_text"), "content": "This is the second step"},
        ]

    @property
    def layout(self) -> html.Div:
        return html.Div(
            children=[
                html.Span(
                    "Here is some blue text to explain... ",
                    id=self.uuid("blue_text"),
                    style={"color": "blue"},
                ),
                html.Span(
                    " ...and here is some red text that also needs an explanation.",
                    id=self.uuid("red_text"),
                    style={"color": "red"},
                ),
            ]
        )



import io
import glob
import shutil
import functools
import hashlib
import inspect
import pathlib
import warnings
from collections import defaultdict
from typing import Callable, List, Union, Any

import pandas as pd
from tqdm import tqdm


class WebvizStorage:

    RETURN_TYPES = [pd.DataFrame, pathlib.Path, io.BytesIO]

    def __init__(self) -> None:
        self._use_storage = False
        self.storage_functions: set = set()
        self.storage_function_argvalues: defaultdict = defaultdict(dict)

    def register_function(self, func: Callable) -> None:
        """This function is automatically called by the function
        decorator @webvizstore, registering the function it decorates.
        """
        return_type = inspect.getfullargspec(func).annotations["return"]

        if return_type not in WebvizStorage.RETURN_TYPES:
            raise NotImplementedError(
                f"Webviz storage type must be one of {WebvizStorage.RETURN_TYPES}"
            )

        self.storage_functions.add(func)

    @property
    def storage_folder(self) -> pathlib.Path:
        return self._storage_folder

    @storage_folder.setter
    def storage_folder(self, path: pathlib.Path) -> None:
        path.mkdir(parents=True, exist_ok=True)
        self._storage_folder = path

    @property
    def use_storage(self) -> bool:
        return self._use_storage

    @use_storage.setter
    def use_storage(self, use_storage: bool) -> None:
        self._use_storage = use_storage

    def register_function_arguments(self, functionarguments: List[tuple]) -> None:
        """The input here is from class functions `add_webvizstore(self)`
        in the different plugins requested from the configuration file.

        The input is as follows:
            [(func1, argumentcombinations), (func2, argumentcombinations), ...]

        where argumentcombinations is a list of kwargs dictionaries.
        """

        for func, arglist in functionarguments:
            undec_func = WebvizStorage._undecorate(func)
            for args in arglist:
                argtuples = WebvizStorage._dict_to_tuples(
                    WebvizStorage.complete_kwargs(undec_func, args)
                )
                if repr(argtuples) not in self.storage_function_argvalues[undec_func]:
                    self.storage_function_argvalues[undec_func][
                        repr(argtuples)
                    ] = argtuples

                for (argname, argvalue) in argtuples:
                    # Check that values of decorated functions are not pandas objects.
                    # This function could probably at some point be moved into a pytest
                    # fixture (given that plugins use type hints) in order to slightly
                    # reduce running time.
                    if isinstance(argvalue, (pd.DataFrame, pd.Series)):
                        warnings.warn(
                            f"{func.__module__}.{func.__name__} is a @webvizstore decorated "
                            f"function, and argument {argname} has been given a pandas "
                            "object as value. Since pandas.DataFrames and pandas.Series are "
                            "known to not have unique/deterministic __repr__ functions, "
                            "they do not work well with @webvizstore (or flask-caching). "
                            "Consider moving to another object with a deterministic "
                            "__repr__ representation more suitable for hashing.",
                            RuntimeWarning,
                            stacklevel=0,
                        )

    def _unique_path(self, func: Callable, argtuples: tuple) -> str:
        """Encodes the argumenttuples as bytes, and then does a sha256 on that.
        Mutable arguments are accepted in the argument tuples, however it is
        the plugin author that needs to be responsible for making sure that
        instances representing different input has different values for
        `__repr__`
        """

        hashed_args = hashlib.sha256(repr(argtuples).encode()).hexdigest()

        filename = f"{func.__module__}-{func.__name__}-{hashed_args}"

        return str(self.storage_folder / filename)

    @staticmethod
    def _undecorate(func: Callable) -> Callable:
        """This unwraps potential multiple level of decorators, to get
        access to the original function.
        """

        while hasattr(func, "__wrapped__"):
            func = func.__wrapped__  # type: ignore[attr-defined]

        return func

    @staticmethod
    def string(func: Callable, kwargs: dict) -> str:
        strkwargs = ", ".join([f"{k}={v!r}" for k, v in kwargs.items()])

        return f"{func.__name__}({strkwargs})"

    @staticmethod
    def _dict_to_tuples(dictionary: dict) -> tuple:
        """Since dictionaries are not hashable, this is a helper function
        converting a dictionary into a sorted tuple."""

        return tuple(sorted(dictionary.items()))

    @staticmethod
    def complete_kwargs(func: Callable, kwargs: dict) -> dict:
        """This takes in a dictionary kwargs, and returns an updated
        dictionary where missing arguments are added with default values."""

        argspec = inspect.getfullargspec(func)

        if argspec.defaults is not None:
            ndef = len(argspec.defaults)
            for arg, val in zip(argspec.args[-ndef:], argspec.defaults):
                if arg not in kwargs:
                    kwargs[arg] = val

        return kwargs

    def get_stored_data(
        self, func: Callable, *args: Any, **kwargs: Any
    ) -> Union[pd.DataFrame, pathlib.Path, io.BytesIO]:

        argspec = inspect.getfullargspec(func)
        for arg_name, arg in zip(argspec.args, args):
            kwargs[arg_name] = arg

        WebvizStorage.complete_kwargs(func, kwargs)
        return_type = inspect.getfullargspec(func).annotations["return"]

        path = self._unique_path(func, WebvizStorage._dict_to_tuples(kwargs))

        try:
            if return_type == pd.DataFrame:
                return pd.read_parquet(f"{path}.parquet")
            if return_type == pathlib.Path:
                return pathlib.Path(glob.glob(f"{path}*")[0])
            if return_type == io.BytesIO:
                return io.BytesIO(pathlib.Path(path).read_bytes())
            raise ValueError(f"Unknown return type {return_type}")

        except OSError as exc:
            raise OSError(
                f"Could not find file {path}, which should be the "
                "stored output of the function call "
                f"{WebvizStorage.string(func, kwargs)}."
            ) from exc

    def build_store(self) -> None:

        total_calls = sum(
            len(calls) for calls in self.storage_function_argvalues.values()
        )

        with tqdm(
            total=total_calls, bar_format="{l_bar} {bar} | Saved {n_fmt}/{total_fmt}"
        ) as progress_bar:
            for func in self.storage_functions:
                if self.storage_function_argvalues[func]:
                    progress_bar.write(
                        f"Storing output of {func.__module__}.{func.__name__}"
                    )
                for argtuples in self.storage_function_argvalues[func].values():
                    output = func(**dict(argtuples))
                    path = self._unique_path(func, argtuples)

                    if isinstance(output, pd.DataFrame):
                        output.to_parquet(f"{path}.parquet")
                    elif isinstance(output, pathlib.Path):
                        shutil.copy(output, f"{path}{output.suffix}")
                    elif isinstance(output, io.BytesIO):
                        pathlib.Path(path).write_bytes(output.getvalue())
                    else:
                        raise ValueError(f"Unknown return type {type(output)}")

                    progress_bar.update()


def webvizstore(func: Callable) -> Callable:

    WEBVIZ_STORAGE.register_function(func)

    @functools.wraps(func)
    def wrapper_decorator(*args: Any, **kwargs: Any) -> Any:
        if WEBVIZ_STORAGE.use_storage:
            return WEBVIZ_STORAGE.get_stored_data(func, *args, **kwargs)
        return func(*args, **kwargs)

    return wrapper_decorator


WEBVIZ_STORAGE = WebvizStorage()


@webvizstore
def get_resource(filename: str) -> pathlib.Path:
    """Utility funtion for getting a filename which works both for
    non-portable and portable webviz instances."""

    return pathlib.Path(filename)



import sys
import json
import time
import secrets
import pathlib
import argparse
import tempfile
import webbrowser
from typing import Dict

import tqdm
import jinja2
import requests

from . import azure_cli
from . import github_cli
from . import radix_cli
from . import interactive_terminal
from .radix_configuration import radix_configuration
from .azure_configuration import azure_configuration


def create_radix_config(
    build_directory: pathlib.Path, settings: Dict[str, str]
) -> None:
    template_environment = jinja2.Environment(  # nosec
        loader=jinja2.PackageLoader("webviz_config", "templates"),
        undefined=jinja2.StrictUndefined,
        autoescape=False,
    )
    template = template_environment.get_template("radixconfig.yaml.jinja2")

    (build_directory / "radixconfig.yaml").write_text(template.render(settings))
    (build_directory / "auth-state.Dockerfile").write_text(
        "FROM redis:alpine\nUSER 999"
    )


def website_online(url: str) -> bool:
    try:
        requests.get(url)
    except requests.exceptions.ConnectionError:
        return False
    return True


def radix_initial_deployment(github_slug: str, build_directory: pathlib.Path) -> None:
    """Creates a Radix configuration file."""
    # pylint: disable=too-many-statements

    if not radix_cli.binary_available():
        raise RuntimeError(
            "The Radix CLI is not available in $PATH. If you haven't downloaded it, "
            "you can find binaries here: https://github.com/equinor/radix-cli/releases"
        )

    if not github_cli.binary_available():
        raise RuntimeError(
            "The GitHub CLI is not available in $PATH. If you haven't downloaded it, "
            "you can find binaries here: https://github.com/cli/cli/releases"
        )

    if not github_cli.logged_in():
        print("You are not logged in with GitHub CLI. Follow instructions below.")
        github_cli.log_in()

    if not radix_cli.logged_in():
        print("You are not logged in with Radix CLI. Follow instructions below.")
        radix_cli.log_in()

    repository_url = f"https://github.com/{github_slug}"
    if github_cli.repo_exists(github_slug):
        raise ValueError(f"GitHub repository {github_slug} already exists.")

    radix_config = radix_configuration()

    azure_configuration_values = azure_configuration()

    interactive_terminal.terminal_title("Initializing automatic deployment")

    progress_bar = tqdm.tqdm(total=15, bar_format="{l_bar} {bar} |")

    progress_bar.write("✓ Configuring Azure app registration")
    azure_configuration_values.update(
        azure_cli.azure_app_registration_setup(
            display_name=azure_configuration_values["display_name"],
            proxy_redirect_url=radix_config["app_url"] + "/oauth2/callback",
        )
    )
    progress_bar.update()

    # Give Azure some seconds to store the enterprise application
    # before opening the link for the user:
    time.sleep(5)
    webbrowser.open(
        "https://portal.azure.com/#blade/Microsoft_AAD_IAM/ManagedAppMenuBlade/Users/"
        f"objectId/{azure_configuration_values['object_id']}/"
        f"appId/{azure_configuration_values['app_registration_id']}"
    )

    progress_bar.write("✓ Uploading Webviz file resources to Azure storage container.")
    azure_cli.storage_container_upload_folder(
        subscription=azure_configuration_values["subscription"],
        resource_group=azure_configuration_values["resource_group"],
        storage_name=azure_configuration_values["storage_account_name"],
        container_name=azure_configuration_values["storage_container_name"],
        source_folder=build_directory / "resources",
    )
    progress_bar.update()

    with tempfile.TemporaryDirectory() as tmp_dir:
        temp_dir = pathlib.Path(tmp_dir)

        progress_bar.write(f"✓ Creating new private repository ({repository_url})")
        cloned_path = github_cli.create_github_repository(
            github_slug, directory=temp_dir
        )
        progress_bar.update()

        progress_bar.write("✓ Adding files to GitHub repository.")
        dict_to_store = {
            "azure": {
                key: value
                for key, value in azure_configuration_values.items()
                if "secret" not in key
            },
            "radix": radix_config,
        }
        (build_directory / "deploy_settings.json").write_text(
            json.dumps(dict_to_store, indent=4, sort_keys=True)
        )
        create_radix_config(
            build_directory,
            {
                "app_registration_id": azure_configuration_values[
                    "app_registration_id"
                ],
                "application_name": radix_config["application_name"],
                "azure_storage_container_name": azure_configuration_values[
                    "storage_container_name"
                ],
                "tenant_id": azure_configuration_values["tenant_id"],
            },
        )
        github_cli.commit_portable_webviz(
            github_slug=github_slug, source_directory=build_directory
        )
        (build_directory / "deploy_settings.json").unlink()
        (build_directory / "radixconfig.yaml").unlink()
        progress_bar.update()

        progress_bar.write(
            "✓ Turning on GitHub repository dependency vulnerability alerts."
        )
        github_cli.turn_on_github_vulnerability_alers(directory=cloned_path)
        progress_bar.update()

        progress_bar.write(
            f"✓ Creating Radix application '{radix_config['application_name']}' "
            f"in {radix_config['context']}."
        )
        webhook_secret = secrets.token_urlsafe()
        public_key = radix_cli.create_application(
            application_name=radix_config["application_name"],
            owner_email=radix_config["owner_email"],
            repository_url=repository_url,
            shared_secret=webhook_secret,
            wbs=radix_config["wbs"],
            context=radix_config["context"],
        )
        progress_bar.update()

        # Wait for application to be reconciled
        time.sleep(5)

        progress_bar.write(
            "✓ Adding webhook from GitHub repository to Radix application."
        )
        github_cli.add_webhook(
            directory=cloned_path,
            receiver_url=radix_config["webhook_receiver_url"],
            secret=webhook_secret,
        )
        progress_bar.update()

        progress_bar.write("✓ Adding Radix deploy key to GitHub repository.")
        github_cli.add_deploy_key(
            directory=cloned_path, title="Radix deploy key", key=public_key
        )
        progress_bar.update()

    # Wait for application to be reconciled
    time.sleep(5)

    progress_bar.write("✓ Initializing build workflow of Radix application.")
    radix_cli.build_and_deploy_application(
        application_name=radix_config["application_name"],
        context=radix_config["context"],
    )
    progress_bar.update()

    progress_bar.write(
        "✓ Setting Radix secrets when Radix has built the application... "
        "please wait (approx. 3-4 min)."
    )
    for key, value in {
        "OAUTH2_PROXY_CLIENT_SECRET": azure_configuration_values["proxy_client_secret"],
        "OAUTH2_PROXY_COOKIE_SECRET": azure_configuration_values["proxy_cookie_secret"],
        "OAUTH2_PROXY_REDIRECT_URL": azure_configuration_values["proxy_redirect_url"],
    }.items():
        radix_cli.set_radix_secret(
            application_name=radix_config["application_name"],
            environment="prod",
            component="auth",
            key=key,
            value=value,
            context=radix_config["context"],
        )
        progress_bar.write(f"  ✓ Radix '{key}' secret set.")
        progress_bar.update()

    for key, value in {
        "main-appstorage-csiazurecreds-accountkey": azure_configuration_values[
            "storage_account_key_secret"
        ],
        "main-appstorage-csiazurecreds-accountname": azure_configuration_values[
            "storage_account_name"
        ],
    }.items():
        radix_cli.set_radix_secret(
            application_name=radix_config["application_name"],
            environment="prod",
            component="main",
            key=key,
            value=value,
            context=radix_config["context"],
        )
        progress_bar.write(f"  ✓ Radix '{key}' secret set.")
        progress_bar.update()

    progress_bar.write("✓ Waiting on Radix application to start and become online.")
    while not website_online(radix_config["app_url"]):
        time.sleep(10)

    progress_bar.update()
    progress_bar.write(
        "✓ Application successfully deployed and online at " + radix_config["app_url"]
    )

    progress_bar.close()

    webbrowser.open(radix_config["app_url"])


def radix_redeploy(github_slug: str, build_directory: pathlib.Path) -> None:

    deploy_settings = json.loads(
        github_cli.read_file_in_repository(github_slug, "deploy_settings.json")
    )
    github_cli.commit_portable_webviz(
        github_slug, build_directory, commit_message="Update"
    )

    azure_cli.storage_container_upload_folder(
        subscription=deploy_settings["azure"]["subscription"],
        resource_group=deploy_settings["azure"]["resource_group"],
        storage_name=deploy_settings["azure"]["storage_account_name"],
        container_name=deploy_settings["azure"]["storage_container_name"],
        source_folder=build_directory / "resources",
    )


def main_radix_deployment(args: argparse.Namespace) -> None:

    if sys.version_info < (3, 8):
        raise RuntimeError("Radix deployment workflow requires at least Python 3.8")

    if not args.portable_app.is_dir():
        raise ValueError(f"{args.portable_app} is not a directory.")

    if args.initial_deploy:
        radix_initial_deployment(args.github_slug, args.portable_app)
    else:
        radix_redeploy(args.github_slug, args.portable_app)



import io
import abc
import base64
import zipfile
import warnings
import sys
import urllib
from uuid import uuid4
from typing import List, Optional, Type, Union, Dict

import bleach
from dash.development.base_component import Component
from dash import Input, Output
import jinja2
import webviz_core_components as wcc

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict


class ZipFileMember(TypedDict):
    filename: str  # Filename to be given within the archive
    content: str  # String representing file content


class EncodedFile(ZipFileMember):
    # Same keys as in ZipFileMember, with the following changes:
    # - filename is now name of the actual downloaded file.
    # - content is now a base64 encoded ASCII string.
    # - mime_type needs to be added as well.
    mime_type: str


def _create_feedback_text(
    plugin_name: str, dist_name: str, dist_version: str, dependencies: Dict[str, str]
) -> str:
    template_environment = jinja2.Environment(  # nosec
        loader=jinja2.PackageLoader("webviz_config", "templates"),
        undefined=jinja2.StrictUndefined,
        autoescape=False,
    )
    template = template_environment.get_template("feedback.md.jinja2")
    return template.render(
        {
            "plugin_name": plugin_name,
            "dist_name": dist_name,
            "dist_version": dist_version,
            "dependencies": dependencies,
        }
    )


class WebvizPluginABC(abc.ABC):
    """All webviz plugins need to subclass this abstract base class,
    e.g.

    ```python
    class MyPlugin(WebvizPluginABC):

        def __init__(self):
            ...

        def layout(self):
            ...
    ```
    """

    # This is the default set of buttons to show in the rendered plugin
    # toolbar. If the list is empty, the subclass plugin layout will be
    # used directly, without any visual encapsulation layout from this
    # abstract base class. The plugins subclassing this abstract base class
    # can override this variable setting by defining a class constant with
    # the same name.
    #
    # Some buttons will only appear if in addition necessary data is available.
    # E.g. download of zip archive will only appear if the plugin also
    # has defined the corresponding callback, and contact person will only
    # appear if the user configuration file has this information.
    TOOLBAR_BUTTONS = [
        "screenshot",
        "expand",
        "contact_person",
        "guided_tour",
        "feedback",
    ]

    # List of plugin specific assets which should be copied
    # over to the ./assets folder in the generated webviz app.
    # This is typically custom JavaScript and/or CSS files.
    # All paths in the returned ASSETS list should be absolute.
    ASSETS: list = []

    def __init__(self, screenshot_filename: str = "webviz-screenshot.png") -> None:
        """If a plugin/subclass defines its own `__init__` function
        (which they usually do), they should remember to call
        ```python
        super().__init__()
        ```
        in its own `__init__` function in order to also run the parent `__init__`.
        """

        self._plugin_uuid = uuid4()
        self._screenshot_filename = screenshot_filename
        self._add_download_button = False

    def uuid(self, element: str) -> str:
        """Typically used to get a unique ID for some given element/component in
        a plugins layout. If the element string is unique within the plugin, this
        function returns a string which is guaranteed to be unique also across the
        application (even when multiple instances of the same plugin is added).

        Within the same plugin instance, the returned uuid is the same for the same
        element string. I.e. storing the returned value in the plugin is not necessary.

        Main benefit of using this function instead of creating a UUID directly,
        is that the abstract base class can in the future provide IDs that
        are consistent across application restarts (i.e. when the webviz configuration
        file changes in a non-portable setting).
        """

        return f"{element}-{self._plugin_uuid}"

    @property
    @abc.abstractmethod
    def layout(self) -> Union[str, Type[Component]]:
        """This is the only required function of a Webviz plugin.
        It returns a Dash layout which by webviz-config is added to
        the main Webviz application.
        """

    @property
    def _plugin_wrapper_id(self) -> str:
        return f"plugin-wrapper-{self._plugin_uuid}"

    @property
    def plugin_data_output(self) -> Output:
        self._add_download_button = True
        return Output(self._plugin_wrapper_id, "download")

    @property
    def plugin_data_requested(self) -> Input:
        return Input(self._plugin_wrapper_id, "data_requested")

    @staticmethod
    def _reformat_tour_steps(steps: List[dict]) -> List[dict]:
        return [
            {"selector": "#" + step["id"], "content": step["content"]} for step in steps
        ]

    @staticmethod
    def plugin_compressed_data(
        filename: str, content: List[ZipFileMember]
    ) -> EncodedFile:
        with io.BytesIO() as bytes_io:
            with zipfile.ZipFile(bytes_io, "w") as zipped_data:
                for data in content:
                    zipped_data.writestr(data["filename"], data["content"])
            return {
                "filename": filename,
                "content": base64.b64encode(bytes_io.getvalue()).decode("ascii"),
                "mime_type": "application/zip",
            }

    @staticmethod
    def plugin_data_compress(content: List[ZipFileMember]) -> EncodedFile:
        warnings.warn(
            "Use 'plugin_compressed_data' instead of 'plugin_data_compress'",
            DeprecationWarning,
        )
        return WebvizPluginABC.plugin_compressed_data("webviz-data.zip", content)

    def _make_extended_deprecation_warnings(
        self,
        plugin_deprecation_warnings: Optional[List[str]] = None,
        argument_deprecation_warnings: Optional[List[str]] = None,
    ) -> List[Dict[str, str]]:
        # pylint: disable=import-outside-toplevel
        from .plugins import PLUGIN_METADATA, PLUGIN_PROJECT_METADATA

        extended_deprecation_warnings: List[Dict[str, str]] = []

        plugin_name = self.__class__.__name__
        dist_name = PLUGIN_METADATA[plugin_name]["dist_name"]
        metadata = PLUGIN_PROJECT_METADATA[dist_name]

        if plugin_deprecation_warnings:
            url = (
                f"{metadata['documentation_url']}/#/plugin_deprecations?id={plugin_name.lower()}"
                if metadata["documentation_url"] is not None
                else ""
            )

            for plugin_deprecation_warning in plugin_deprecation_warnings:
                extended_deprecation_warnings.append(
                    {"message": plugin_deprecation_warning, "url": url}
                )
        if argument_deprecation_warnings:
            url = (
                f"{metadata['documentation_url']}/#/argument_deprecations?id={plugin_name.lower()}"
                if metadata["documentation_url"] is not None
                else ""
            )

            for argument_deprecation_warning in argument_deprecation_warnings:
                extended_deprecation_warnings.append(
                    {"message": argument_deprecation_warning, "url": url}
                )
        return extended_deprecation_warnings

    def _make_feedback_url(self) -> str:
        # pylint: disable=import-outside-toplevel
        from .plugins import PLUGIN_METADATA, PLUGIN_PROJECT_METADATA

        plugin_name = self.__class__.__name__
        dist_name = PLUGIN_METADATA[plugin_name]["dist_name"]
        metadata = PLUGIN_PROJECT_METADATA[dist_name]

        feedback_url = ""
        if metadata["tracker_url"] is not None:
            queries = {
                "title": "",
                "body": _create_feedback_text(
                    plugin_name,
                    dist_name,
                    metadata["dist_version"],
                    metadata["dependencies"],
                ),
            }

            feedback_url = (
                f"{metadata['tracker_url']}/new?{urllib.parse.urlencode(queries)}"
                if "github.com" in metadata["tracker_url"]
                else metadata["tracker_url"]
            )

        return feedback_url

    def plugin_layout(
        self,
        contact_person: Optional[dict] = None,
        plugin_deprecation_warnings: Optional[List[str]] = None,
        argument_deprecation_warnings: Optional[List[str]] = None,
    ) -> Union[str, Type[Component]]:
        """This function returns (if the class constant SHOW_TOOLBAR is True,
        the plugin layout wrapped into a common webviz config plugin
        component, which provides some useful buttons like download of data,
        show data contact person and download plugin content to png.

        CSV download button will only appear if the plugin class has a property
        `csv_string` which should return the appropriate csv data as a string.

        If `TOOLBAR_BUTTONS` is empty, this functions returns the same
        dash layout as the plugin class provides directly.
        """

        buttons = self.__class__.TOOLBAR_BUTTONS.copy()

        if contact_person is None:
            contact_person = {}
        else:
            # Sanitize the configuration user input
            for key in contact_person:
                contact_person[key] = bleach.clean(str(contact_person[key]))

        if self._add_download_button:
            buttons.append("download")

        if buttons or plugin_deprecation_warnings or argument_deprecation_warnings:
            # pylint: disable=no-member
            return wcc.WebvizPluginPlaceholder(
                id=self._plugin_wrapper_id,
                buttons=buttons,
                contact_person=contact_person,
                children=[self.layout],
                screenshot_filename=self._screenshot_filename,
                tour_steps=WebvizPluginABC._reformat_tour_steps(
                    self.tour_steps  # type: ignore[attr-defined]
                )
                if "guided_tour" in buttons and hasattr(self, "tour_steps")
                else [],
                deprecation_warnings=self._make_extended_deprecation_warnings(
                    plugin_deprecation_warnings, argument_deprecation_warnings
                ),
                feedback_url=self._make_feedback_url(),
            )

        return self.layout
